#include <ini.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "ctx.h"
#include "error.h"
#include "intl.h"
#include "log.h"
#include "util.h"

struct __config_args {
  config_t *config;
  ctx_t    *ctx;
};

bool config_contains_target(config_t *config, char *name) {
  target_t *cur = config->t_last;
  while (cur) {
    if (eq(cur->name, name))
      return true;
    cur = cur->next;
  }
  return false;
}

bool config_is_valid(config_t *config) {
  if (NULL == config->name) {
    error_set(_("Configuration does not have a name field"));
    return false;
  }

  if (NULL == config->author) {
    error_set(_("Configuration does not have an author field"));
    return false;
  }

  if (NULL == config->t_last) {
    error_set(_("Configuration does not have any targets"));
    return false;
  }

  target_t *cur = config->t_first;
  while (cur) {
    if (!target_is_valid(cur))
      return false;
    cur = cur->next;
  }

  return true;
}

bool config_add_target(config_t *config, char *name) {
  target_t *new = malloc(sizeof(target_t));
  target_init(new);
  new->name = name;

  config->t_len++;

  if (NULL == config->t_first || NULL == config->t_last) {
    config->t_first = new;
    config->t_last  = new;
    return true;
  }

  config->t_last->next = new;
  config->t_last       = new;
  return true;
}

int config_handler(void *data, const char *section, const char *key, const char *value) {
#define MATCH(s, k) eq((char *)section, s) && eq((char *)key, k)
#define MATCHKEY(k) eq((char *)key, k)

  if (strlen(section) == 0)
    return 1;

  struct __config_args *args   = data;
  config_t             *config = args->config;
  ctx_t                *ctx    = args->ctx;

  if (MATCH("details", "name")) {
    config->name = strdup(value);
    return 1;
  } else if (MATCH("details", "author")) {
    config->author = strdup(value);
    return 1;
  } else if (MATCH("details", "keywords")) {
    clist_from_str(&config->keywords, (char *)value);
    return 1;
  } else if (MATCH("details", "desc")) {
    config->desc = strdup(value);
    return 1;
  } else if (eq((char *)section, "details")) {
    goto UNKNOWN;
    return 1;
  }

  if (NULL == config->t_last || (NULL != config->t_last->name && !eq(config->t_last->name, (char *)section))) {
    if (config_contains_target(config, (char *)section)) {
      error_set(_("Configuration contains multiple targets with the same name"));
      errno = ConfigMultiple;
      return 0;
    }
    config_add_target(config, strdup(section));
  }

  if (MATCHKEY("desc")) {
    config->t_last->desc = strdup(value);
    return 1;
  } else if (MATCHKEY("dst")) {
    int len             = strlen(value) + strlen(ctx->home_path) + 2;
    config->t_last->dst = malloc(len);
    join(config->t_last->dst, ctx->home_path, value);
    return 1;
  } else if (MATCHKEY("src")) {
    config->t_last->src = strdup(value);
    return 1;
  } else if (MATCHKEY("require")) {
    clist_from_str(&config->t_last->require, strdup(value));
    return 1;
  } else if (MATCHKEY("nogen")) {
    config->t_last->nogen = eq((char *)value, "true");
    return 1;
  }

UNKNOWN:
  error_set(_("Key %s is unknown"), key);
  errno = ConfigUnknown;
  return 0;
}

void config_free(config_t *config) {
  free(config->name);
  free(config->desc);
  free(config->author);

  target_t *cur = config->t_first, *prev = NULL;
  while (cur != NULL) {
    prev = cur;
    cur  = cur->next;

    target_free(prev);
    free(prev);
  }

  clist_free(&config->keywords);

  config->name    = NULL;
  config->desc    = NULL;
  config->author  = NULL;
  config->t_first = NULL;
  config->t_last  = NULL;
  config->t_len   = 0;
  clist_init(&config->keywords);
}

bool config_load(ctx_t *ctx, config_t *config, char *path) {
  if (!exists(path) || is_dir(path)) {
    error_set(_("Configuration file not found"));
    errno = ConfigNotFound;
    return false;
  }

  bzero(config, sizeof(config_t));
  clist_init(&config->keywords);

  struct __config_args args = {
      .config = config,
      .ctx    = ctx,
  };

  if (ini_parse(path, config_handler, &args) < 0) {
    config_free(config);
    if (errno != ConfigUnknown || errno != ConfigMultiple) {
      error_set(_("Failed to parse configuration file"));
      errno = ConfigParseFail;
    }
    return false;
  }

  if (!config_is_valid(config)) {
    errno = ConfigInvalid;
    config_free(config);
    return false;
  }

  return true;
}

void config_print(config_t *config) {
  info(_("Configuration details:\n"));
  printf(FG_BOLD "    %s     " FG_RESET "=> %s\n", _("Name"), config->name);
  if (NULL != config->desc)
    printf(FG_BOLD "    %s     " FG_RESET "=> %s\n", _("Desc"), config->desc);
  printf(FG_BOLD "    %s   " FG_RESET "=> %s\n", _("Author"), config->author);

  printf(FG_BOLD "    %s " FG_RESET "=> ", _("Keywords"));
  for (int i = 0; i < config->keywords.s; i++)
    printf("%s ", config->keywords.c[i]);
  printf("\n\n");
}
