// clang-format off

/*

 *  mc | MatterLinux Configuration Manager
 *  MatterLinux 2023-2024 (https://matterlinux.xyz)

 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.

 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/

// clang-format on 

#include <libmp/util.h>
#include <locale.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "args.h"
#include "cmd.h"
#include "ctx.h"
#include "error.h"
#include "intl.h"
#include "lock.h"
#include "log.h"
#include "paths.h"
#include "util.h"

ctx_t ctx;

void handler(int sig){
  unlock(ctx.lock_path);
  printf("\e[?25h\n");
  if(SIGSEGV == sig)
    error(_("Program received a segfault"));
  exit(1);
}

int main(int argc, char *argv[]) {
  signal(SIGINT, handler);
  signal(SIGSEGV, handler);

  setbuf(stdout, NULL);
  setlocale(LC_ALL, "");
  textdomain("confer");

  cmd_t commands[] = {
      {.name = "pull", .lock = true,   .desc = _("pull down a configuration"), .func = cmd_pull   },
      {.name = "gen",  .lock = false,  .desc = _("build the configuration in the current directory"),  .func = cmd_gen   },
  };
  args_t  *args = NULL;
  char *homedir = NULL;
  bool ret = false;

  args = args_parse(argc, argv);
  if (args->count <= 0 || args->list[0].name != NULL)
    goto help;

  if((homedir = args_get_string(args, "home")) == NULL)
    homedir = getenv("HOME");

  if(!ctx_init(&ctx, homedir)){
    switch (errno) {
    case HomedirInvalid:
      error(_("Please specify the home directory with "FG_BOLD"--home"FG_RESET" option or specify the "FG_BOLD"HOME"FG_RESET" environment variable"));
      break;

    case HomedirAccessFail:
      error(_("Cannot access to the home directory ("FG_BOLD"%s"FG_RESET")"), homedir);
      break;
    }

    goto end;
  }

  for (int i = 0; i < sizeof(commands) / sizeof(cmd_t); i++) {
    if (!eq(commands[i].name, args->list[0].value))
      continue;

    if (commands[i].lock) {
      switch (lock(ctx.lock_path)) {
      case ALREADY_LOCKED:
        error(_("Failed to lock, confer is already running"));
        goto end;

      case LOCK_ERROR:
        error(_("Failed to lock, do you have access to the home directory?"));
        goto end;

      default:
        break;
      }
    }

    ret = commands[i].func(&ctx, args);
    goto end;
  }

  error(_("Command not found: " FG_BOLD "%s"), args->list[0].value);
help:
  info(_("MatterLinux configuration manager (version %s)"), VERSION);
  info(_("Usage: " FG_BOLD "%s [command] <options> <arguments>"), argv[0]);
  printf("\n");
  info(_("Here is a list of available commands:"));

  for (int i = 0; i < sizeof(commands) / sizeof(cmd_t); i++)
    printf("    " FG_BOLD "%s" FG_RESET ":\t %s\n", commands[i].name, commands[i].desc);
  printf("\n");

  info(_("Here is a list of available options:"));

  printf(
      _("    " FG_BOLD "--yes" FG_RESET ":\t skip the confirmation\n"));
  printf(_("    " FG_BOLD "--home" FG_RESET ":\t path to your home directory\n\n"));

  info(_("Licensed under GPLv3, see https://www.gnu.org/licenses/ for more information"));

  ret = true;

end:
  unlock(ctx.lock_path);
  args_free(args);
  ctx_free(&ctx);

  return ret ? EXIT_SUCCESS : EXIT_FAILURE;
}
